import unittest
from pathlib import Path
import pandas as pd
import lusid

from finbourne_sdk_utils.cocoon.seed_sample_data import seed_data
from finbourne_sdk_utils.cocoon.utilities import load_json_file
from finbourne_sdk_utils.cocoon.utilities import create_scope_id
import logging
import datetime
import pytz
import random
from unittest.mock import Mock

logger = logging.getLogger()

secrets_file = Path(__file__).parent.parent.parent.joinpath("secrets.json")
seed_sample_data_override_csv = Path(__file__).parent.joinpath(
    "data/seed_sample_data/sample_data_override.csv"
)
sample_data_csv = Path(__file__).parent.joinpath(
    "data/seed_sample_data/sample_data.csv"
)


class CocoonSeedDataTestsBase(object):

    """
    Class description:
    ------------------
    This class creates the tests we use in other classes below.
    We split the file loading and tests into separate classes to improve readability and reuse.
    This class can be extended to your own tests as follows:
    (1) Create a new class with this class as a subclass.
    (2) Load data into LUSID via the setup of new class.
    (3) Run tests against new data.
    """

    def test_transactions_from_response(self):
        transactions_from_response = self.api_factory.build(
            lusid.api.TransactionPortfoliosApi
        ).get_transactions(
            scope=self.scope,
            code=self.sample_data["portfolio_code"].to_list()[0],
            property_keys=[f"Transaction/{self.scope}/strategy"],
        )

        self.assertEqual(
            set([txn.transaction_id for txn in transactions_from_response.values]),
            set(self.sample_data["txn_id"].to_list()),
        )

    def test_portfolio_from_response(self):
        portfolio_from_response = self.api_factory.build(
            lusid.api.PortfoliosApi
        ).get_portfolio(
            scope=self.scope, code=self.sample_data["portfolio_code"].to_list()[0]
        )

        self.assertEqual(
            portfolio_from_response.id.code,
            self.sample_data["portfolio_code"].to_list()[0],
        )

    def test_bad_file_type(self):
        with self.assertRaises(ValueError) as error:
            seed_data(
                self.api_factory,
                ["bad_file_type"],
                self.scope,
                seed_sample_data_override_csv,
                file_type="csv",
            )

        self.assertEqual(
            error.exception.args[0],
            "The provided file_type of bad_file_type has no associated mapping",
        )

    def test_instruments_from_response(self):
        # In this test we take a random sample from the dataframe.
        # We need to make a seperate get_instrument request for each item in the sample, so the test might become
        # inefficient if we take the entire dataframe.
        # A sample of approximately 10 should prove the function works as expected.

        instruments_api = self.api_factory.build(lusid.api.InstrumentsApi)

        random_10_ids = set(
            random.choices(self.sample_data["instrument_id"].to_list(), k=10)
        )
        response_from_random_10 = set(
            [
                response.identifiers["ClientInternal"]
                for response in [
                    instruments_api.get_instrument(
                        identifier_type="ClientInternal", identifier=id
                    )
                    for id in random_10_ids
                ]
            ]
        )

        self.assertEqual(random_10_ids, response_from_random_10)

    def test_sub_holding_keys(self):
        holdings_response = self.api_factory.build(
            lusid.api.TransactionPortfoliosApi
        ).get_holdings(
            scope=self.scope, code=self.sample_data["portfolio_code"].to_list()[0],
        )

        list_of_prop_values = [
            holding.sub_holding_keys[f"Transaction/{self.scope}/strategy"].value
            for holding in holdings_response.values
        ]
        unique_strategy_labels = set([prop.label_value for prop in list_of_prop_values])

        self.assertEqual(
            unique_strategy_labels, set(self.sample_data["strategy"].to_list())
        )


class CocoonTestSeedDataNoMappingOverrideCSV(
    unittest.TestCase, CocoonSeedDataTestsBase
):
    @classmethod
    def setUpClass(cls) -> None:

        cls.scope = create_scope_id().replace("-", "_")
        cls.api_factory = lusid.SyncApiClientFactory( )
        cls.sample_data = pd.read_csv(sample_data_csv)

        cls.domain_list = ["portfolios", "instruments", "transactions"]

        cls.seed_data = seed_data(
            cls.api_factory,
            cls.domain_list,
            cls.scope,
            sample_data_csv,
            sub_holding_keys=[f"Transaction/{cls.scope}/strategy"],
            file_type="csv",
        )

    @classmethod
    def tearDownClass(cls) -> None:
        # Delete portfolio once tests are concluded
        cls.api_factory.build(lusid.PortfoliosApi).delete_portfolio(
            cls.scope, cls.sample_data["portfolio_code"].to_list()[0]
        )

    def test_return_dict(self):

        seed_data = self.seed_data

        self.assertEqual(type(seed_data), dict)
        self.assertEqual(set(seed_data.keys()), set(self.domain_list))


class CocoonTestSeedDataWithMappingOverrideCSV(
    unittest.TestCase, CocoonSeedDataTestsBase
):
    @classmethod
    def setUpClass(cls) -> None:
        cls.scope = create_scope_id().replace("-", "_")
        cls.api_factory = lusid.SyncApiClientFactory( )
        
        cls.sample_data = pd.read_csv(seed_sample_data_override_csv)

    @classmethod
    def tearDownClass(cls) -> None:
        # Delete portfolio once tests are concluded
        cls.api_factory.build(lusid.PortfoliosApi).delete_portfolio(
            cls.scope, cls.sample_data["portfolio_code"].to_list()[0]
        )

    def test_override_with_custom_mapping(self):
        result = seed_data(
            self.api_factory,
            ["portfolios", "instruments", "transactions"],
            self.scope,
            seed_sample_data_override_csv,
            mappings=dict(
                load_json_file(
                    Path(__file__).parent.parent.parent.joinpath(
                        "integration",
                        "cocoon",
                        "data",
                        "seed_sample_data",
                        "seed_sample_data_override.json",
                    )
                )
            ),
            sub_holding_keys=[f"Transaction/{self.scope}/strategy"],
            file_type="csv",
        )
        self.assertEqual(len(result["portfolios"][0]["portfolios"]["success"]), 1)
        self.assertEqual(len(result["instruments"][0]["instruments"]["success"]), 1)
        self.assertEqual(len(result["transactions"][0]["transactions"]["success"]), 1)


class CocoonTestSeedDataNoMappingOverrideExcel(
    unittest.TestCase, CocoonSeedDataTestsBase
):
    @classmethod
    def setUpClass(cls) -> None:
        cls.scope = create_scope_id().replace("-", "_")
        cls.api_factory = lusid.SyncApiClientFactory( )
        cls.sample_data = pd.read_excel(
            Path(__file__).parent.joinpath(
                "data/seed_sample_data/sample_data_excel.xlsx"
            ),
            engine="openpyxl",
        )

        seed_data(
            cls.api_factory,
            ["portfolios", "instruments", "transactions"],
            cls.scope,
            Path(__file__).parent.joinpath(
                "data/seed_sample_data/sample_data_excel.xlsx"
            ),
            sub_holding_keys=[f"Transaction/{cls.scope}/strategy"],
            file_type="xlsx",
        )

    @classmethod
    def tearDownClass(cls) -> None:
        # Delete portfolio once tests are concluded
        cls.api_factory.build(lusid.PortfoliosApi).delete_portfolio(
            cls.scope, cls.sample_data["portfolio_code"].to_list()[0]
        )


class CocoonTestSeedDataUnsupportedFile(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.scope = create_scope_id().replace("-", "_")
        cls.api_factory = Mock()

    def test_bad_file_type(self):
        with self.assertRaises(ValueError) as error:
            seed_data(
                self.api_factory,
                ["portfolios", "instruments", "transactions"],
                self.scope,
                Path(__file__).parent.joinpath("data/seed_sample_data/sample_data.xml"),
                file_type="xml",
            )

        self.assertEqual(
            error.exception.args[0],
            "Unsupported file type, please upload one of the following: ['csv', 'xlsx']",
        )

    def test_inconsistent_file_extensions(self):

        self.file_extension = "xlsx"

        with self.assertRaises(ValueError) as error:
            seed_data(
                self.api_factory,
                ["portfolios", "instruments", "transactions"],
                self.scope,
                seed_sample_data_override_csv,
                file_type=self.file_extension,
            )

        self.assertEqual(
            error.exception.args[0],
            f"""Inconsistent file and file extensions passed: {seed_sample_data_override_csv} does not have file extension {self.file_extension}""",
        )

    def test_file_not_exist(self):

        self.transaction_file = "data/seed_sample_data/file_not_exist.csv"

        self.assertRaises(
            FileNotFoundError,
            seed_data,
            self.api_factory,
            ["portfolios", "instruments", "transactions"],
            self.scope,
            self.transaction_file,
            file_type="csv",
        )


class CocoonTestSeedDataPassDataFrame(unittest.TestCase, CocoonSeedDataTestsBase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.scope = create_scope_id().replace("-", "_")
        cls.api_factory = lusid.SyncApiClientFactory()
        
        cls.test_dataframe = pd.read_csv(sample_data_csv)
        cls.sample_data = pd.read_csv(sample_data_csv)

        seed_data(
            cls.api_factory,
            ["portfolios", "instruments", "transactions"],
            cls.scope,
            cls.test_dataframe,
            sub_holding_keys=[f"Transaction/{cls.scope}/strategy"],
            file_type="DataFrame",
        )

    @classmethod
    def tearDownClass(cls) -> None:
        # Delete portfolio once tests are concluded
        cls.api_factory.build(lusid.PortfoliosApi).delete_portfolio(
            cls.scope, cls.sample_data["portfolio_code"].to_list()[0]
        )



import pandas as pd
import re
import urllib.parse

from finbourne_sdk_utils.lpt.either import Either

rexp = re.compile(r".*page=([^=']{10,}).*")


def page_all_results(fetch_page, page_handler):
    results = []

    def got_page(result):
        results.append(page_handler(result))

        links = [l for l in result.content.links if l.relation == "NextPage"]

        if len(links) > 0:
            match = rexp.match(links[0].href)
            if match:
                return urllib.parse.unquote(match.group(1))
        return None

    page = Either(None)
    while True:
        page = fetch_page(page.right).bind(got_page)
        if page.is_left():
            return page
        if page.right == None:
            break

    return pd.concat(results, ignore_index=True, sort=False)



import os
import unittest
import pandas as pd
from finbourne_sdk_utils import cocoon as cocoon
import lusid

class CocoonTestsSimpleInstruments(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:

        cls.api_factory = lusid.SyncApiClientFactory()

    def test_load_simple_instrument_properties(self,):
         
        data_frame = pd.DataFrame(
            {
                "Security" : [
                    "ALPHABET INC CAP STK USD0.001 CL C",
                    "ALPHABET INC CAPITAL STOCK USD0.001 CL A",
                    "WALT DISNEY CO",
                    "HOME DEPOT INC., COMMON STOCK, $.05 PAR",
                    "MC DONALDS CORP. COMMON STOCK, NO PAR",
                    "NIKE INC CL B"
                ],
                "Ticker"	: [
                    "GOOG",
                    "GOOGL",
                    "DIS",
                    "HD",
                    "MCD",
                    "NKE"
                ],
                "Asset Type" : [
                    "Equity",
                    "Equity",
                    "Equity",
                    "Equity",
                    "Equity",     
                    "Equity",     
                ],
                "Mkt Price" : [
                    150.01,
                    148.66,
                    88.4,
                    370.87,
                    291.27,
                    78.09            
                ],
                "Cusip" : [
                    "02079K107",
                    "02079K305",
                    "254687106",
                    "437076102",
                    "580135101",
                    "654106103"           
                ],
                "Sector" : [
                    "Communication Services",
                    "Communication Services",
                    "Communication Services",
                    "Consumer Discretionary",
                    "Consumer Discretionary",
                    "Consumer Discretionary"            
                ],
                "Yield" : [
                    0.0053,
                    0.0054,
                    0.0085,
                    0.0243,
                    0.0229,
                    0.019           
                ],
                "Level 1" : [
                    "Risk Assets",
                    "Risk Assets",
                    "Risk Assets",
                    "Risk Assets",
                    "Risk Assets",
                    "Risk Assets"
                ],
                "Level 2" : [
                    "Public Equity",
                    "Public Equity",
                    "Public Equity",
                    "Public Equity",
                    "Public Equity",
                    "Public Equity"
                ],
                "Level 3" : [
                    "US Equity",
                    "US Equity",
                    "US Equity",
                    "US Equity",
                    "US Equity",
                    "US Equity"
                ]
            }
        )


        scope = "wealth"
        instrument_mapping = {
            "required": {
                "name": "Security",
                "definition.instrument_type": "$SimpleInstrument",
                "definition.dom_ccy": "$USD",
                "definition.asset_class": "$Equities",
                "definition.simple_instrument_type": "Asset Type"
            },
            "optional": {},
            "identifiers": {
                "ClientInternal": "Ticker",
                "Ticker": "Ticker",
                "Cusip": "Cusip"
            },
            "properties": [
                {"source": "Level 1", "target": "Level 1"},
                {"source": "Level 2", "target": "Level 2"},
                {"source": "Level 3", "target": "Level 3"},
                {"source": "Yield", "target": "Yield"},
                {"source": "Sector", "target": "Sector"}
            ]
        }


        responses = cocoon.cocoon.load_from_data_frame(
            api_factory=self.api_factory,
            scope=scope,
            data_frame=data_frame,
            mapping_required=instrument_mapping["required"],
            mapping_optional=instrument_mapping["optional"],
            identifier_mapping=instrument_mapping["identifiers"],
            file_type="instruments",
            instrument_scope=scope,
            property_columns=instrument_mapping["properties"]
        )

        self.assertEqual(len(responses["instruments"]["success"]), 1)
        self.assertEqual(len(responses["instruments"]["errors"]), 0)
     



import datetime
import importlib
import inspect
import json
import os

from . import lpt
from .either import Either
from .record import Rec
from lusid.api_response import ApiResponse


def ApiConverter(connection, swagger_api):
    def check(f):
        return (inspect.isfunction(f) and f.__name__.endswith("_with_http_info")) or (
            inspect.ismethod(f) and f.__name__.endswith("_with_http_info")
        )

    d = {}

    for api in [i[1] for i in swagger_api.__dict__.items() if i[0].endswith("Api")]:
        jlh = api(connection)
        for n, v in inspect.getmembers(jlh, predicate=check):
            d[n[:-15]] = v
    f = Rec(**d)

    return f


class ExtendedAPI:
    # Constructor
    def __init__(self, config, api, lusid, custom_headers=None):
        class dummy(Exception):
            pass

        self.api = ApiConverter(api, lusid)
        self.models = lusid.models
        self.lusid = lusid
        self.stats_file = config.get("stats", "")
        self.stats = [] if self.stats_file != "" else None

        # See if the api contains the ErrorResponseException

        exc = (
            lusid.rest.ApiException if "ApiException" in lusid.rest.__dict__ else dummy
        )

        self.call = Caller(self.api, self.stats, exc)
        self.call.custom_headers = custom_headers
        self.call.as_at = config.get("asat")

    # Create array of objects using the models
    def from_df(self, df, model, related=None):
        return lpt.from_df(df, model, self.models.__dict__, related)

    def dump_stats(self):
        if self.stats != None:
            lpt.dump_stats(
                self.stats_file,
                self.stats,
                [
                    "startTime",
                    "endTime",
                    "name",
                    "requestId",
                    "duration",
                    "elapsed",
                    "status",
                ],
            )


# Wrapper class to call an API function returning the stats
class Caller:
    def __init__(self, api, stats, exceptionClass):
        self.api = api
        self.stats = stats
        self.exceptionClass = exceptionClass
        self.custom_headers = None
        self.as_at = None

    def __getattr__(self, name):
        fn = getattr(self.api, name)

        # Function that will call the target function
        # returns an Either
        def callApiFn(*args, **kwargs):
            # Add the ' parameter and custom headers
            # adjKwargs = dict(raw=True,custom_headers=self.custom_headers)
            # adjKwargs.update(kwargs)
            adjKwargs = dict([v for v in dict(kwargs).items() if v[1] != None])

            # Add the as_at parameter if provided and valid
            if self.as_at:  # and signature(fn).parameters.get('as_at'):
                adjKwargs.update({"as_at": lpt.to_date(self.as_at)})

            # Measure execution time of the call
            startTime = datetime.datetime.now()
            try:
                result = fn(*args, **(adjKwargs))
                request_id = result.headers.get("lusid-meta-requestId", "n/a")
            except self.exceptionClass as err:
                data = {} if err.body == "" or err.body == b"" else json.loads(err.body)

                instance = data.get("instance", "n/a")
                s = instance.split("insights/logs/")
                request_id = "n/a" if len(s) != 2 else s[1]

                result = ApiResponse(
                    data = Rec(
                        status=err.status,
                        reason=err.reason,
                        code=data.get("code", "n/a"),
                        message=data.get("title", "n/a"),
                        detailed_message=data.get("detail", "n/a"),
                        items=data.get("errorDetails", []),
                        instance=instance,
                        requestId=request_id,
                    ),
                    status_code = err.status,
                    headers={},
                )

            endTime = datetime.datetime.now()

            statistics = Rec(
                name=name,
                startTime=startTime,
                endTime=endTime,
                duration=(endTime - startTime).total_seconds(),
                elapsed=float(result.headers.get("lusid-meta-duration", 0)) / 1000,
                status=result.status_code,
                requestId=request_id,
            )

            if self.stats != None:
                self.stats.append(statistics)

            # If successful, return the output as a 'right'
            if result.headers.get("lusid-meta-success", "False") == "True":
                return Either.Right(Rec(stats=statistics, content=result.data,))
            # Otherwise return a 'left' (a failure)
            return Either.Left(result.data)

        return callApiFn


# Connect to appropriate LUSID environment
def connect(*args, **kwargs):
    # start with no overrides
    overrides = {}

    # include dictionary types
    for i in args:
        overrides.update(i.__dict__)

    # and keyword overrides
    overrides.update(kwargs)

    # Get basic settings from the secrets file
    try:
        with open(overrides.get("secrets", "secrets.json"), "r") as secrets:
            settings = json.load(secrets)
            # apply the overrides
            settings.update(overrides)
    except:
        settings = overrides  # connection may not require secrets. Don't fail (yet)

    # import the appropriate environment for the connection
    env = settings.get("env", ["lusid"])
    connect = importlib.import_module(".connect_{}".format(env[0]), "finbourne_sdk_utils.lpt")

    api = ExtendedAPI(settings, *connect.connect(settings))

    # For debugging purposes, this will display
    if os.getenv("LSE_DEBUG_BODY", "N") == "Y":

        def serializer(standard):
            def fn(request, model):
                s = standard(request, model)
                print(s)
                return s

            return fn

        api.api._serialize.body = serializer(api.api._serialize.body)

    return api



import logging
import os
import unittest
from finbourne_sdk_utils import logger
from datetime import datetime
from parameterized import parameterized
import pytz

from finbourne_sdk_utils.extract.group_holdings import _join_holdings
from lusid.models import (
    PortfolioHolding,
    ModelProperty,
    PropertyValue,
    CurrencyAndAmount,
)

now = datetime.now(pytz.UTC)


class CocoonExtractGroupHoldingsTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.logger = logger.LusidLogger(os.getenv("FBN_LOG_LEVEL", "info"))

    class PortfolioHoldingTemplate(PortfolioHolding):
        """
        This class provides the ability for overriding the template as you see fit
        """

        def __init__(self, **override_init_kwargs):

            default_init_kwargs = {
                "instrument_uid": "LUID_12345678",
                "properties": {
                    "MyProperty": ModelProperty(
                        key="MyProperty", value=PropertyValue(label_value="blah")
                    )
                },
                "holding_type": "B",
                "units": 100,
                "settled_units": 100,
                "cost": CurrencyAndAmount(amount=10000, currency="AUD"),
                "cost_portfolio_ccy": CurrencyAndAmount(amount=0, currency="AUD"),
            }

            init_kwargs = {**default_init_kwargs, **override_init_kwargs}

            super().__init__(**init_kwargs)

    @parameterized.expand(
        [
            [
                "Standard joining preserving portfolio separation",
                {
                    "MyScope : PortfolioA": [PortfolioHoldingTemplate()],
                    "MyScope : PortfolioB": [PortfolioHoldingTemplate()],
                },
                True,
                None,
                {
                    "MyScope : PortfolioA": [PortfolioHoldingTemplate()],
                    "MyScope : PortfolioB": [PortfolioHoldingTemplate()],
                },
            ],
            [
                "Standard joining with no portfolio separation",
                {
                    "MyScope : PortfolioA": [PortfolioHoldingTemplate()],
                    "MyScope : PortfolioB": [PortfolioHoldingTemplate()],
                },
                False,
                "PortGroupScope : MyPortGroup",
                {
                    "PortGroupScope : MyPortGroup": [
                        PortfolioHoldingTemplate(
                            **{
                                "units": 200,
                                "settled_units": 200,
                                "cost": CurrencyAndAmount(amount=20000, currency="AUD"),
                                "properties": {},
                            }
                        )
                    ]
                },
            ],
            [
                "Standard joining with no portfolio separation with Instrument property",
                {
                    "MyScope : PortfolioA": [
                        PortfolioHoldingTemplate(
                            **{
                                "properties": {
                                    "Instrument/default/Name": ModelProperty(
                                        key="Instrument/default/Name",
                                        value=PropertyValue(label_value="Apple"),
                                    )
                                }
                            }
                        )
                    ],
                    "MyScope : PortfolioB": [PortfolioHoldingTemplate()],
                },
                False,
                "PortGroupScope : MyPortGroup",
                {
                    "PortGroupScope : MyPortGroup": [
                        PortfolioHoldingTemplate(
                            **{
                                "units": 200,
                                "settled_units": 200,
                                "cost": CurrencyAndAmount(amount=20000, currency="AUD"),
                                "properties": {
                                    "Instrument/default/Name": ModelProperty(
                                        key="Instrument/default/Name",
                                        value=PropertyValue(label_value="Apple"),
                                    )
                                },
                            }
                        )
                    ]
                },
            ],
            [
                "All merged with same instrument in different currencies",
                {
                    "MyScope : PortfolioA": [
                        PortfolioHoldingTemplate(
                            **{"cost": CurrencyAndAmount(amount=1500, currency="USD")}
                        )
                    ],
                    "MyScope : PortfolioB": [PortfolioHoldingTemplate()],
                },
                False,
                "PortGroupScope : MyPortGroup",
                {
                    "PortGroupScope : MyPortGroup": [
                        PortfolioHoldingTemplate(
                            **{
                                "cost": CurrencyAndAmount(amount=1500, currency="USD"),
                                "properties": {},
                            }
                        ),
                        PortfolioHoldingTemplate(**{"properties": {}}),
                    ]
                },
            ],
            [
                "2x Instrument A, 1x Instrument B",
                {
                    "MyScope : PortfolioA": [
                        PortfolioHoldingTemplate(
                            **{
                                "cost": CurrencyAndAmount(amount=1500, currency="USD"),
                                "instrument_uid": "LUID_ABD29433",
                            }
                        ),
                        PortfolioHoldingTemplate(),
                    ],
                    "MyScope : PortfolioB": [PortfolioHoldingTemplate()],
                },
                False,
                "PortGroupScope : MyPortGroup",
                {
                    "PortGroupScope : MyPortGroup": [
                        PortfolioHoldingTemplate(
                            **{
                                "cost": CurrencyAndAmount(amount=1500, currency="USD"),
                                "instrument_uid": "LUID_ABD29433",
                                "properties": {},
                            }
                        ),
                        PortfolioHoldingTemplate(
                            **{
                                "units": 200,
                                "settled_units": 200,
                                "cost": CurrencyAndAmount(amount=20000, currency="AUD"),
                                "properties": {},
                            }
                        ),
                    ]
                },
            ],
        ]
    )
    def test_join_holdings(
        self, _, holdings_to_join, group_by_portfolio, dict_key, expected_outcome
    ):

        joined_holdings = _join_holdings(
            holdings_to_join=holdings_to_join,
            group_by_portfolio=group_by_portfolio,
            dict_key=dict_key,
        )

        logging.info(joined_holdings)
        logging.info(expected_outcome)

        self.assertEqual(joined_holdings, expected_outcome)
